"""Compute the truth value of x1 OR x2 element-wise."""
import numpy
import numpoly

from ..dispatch import implements, simple_dispatch


@implements(numpy.logical_or)
def logical_or(x1, x2, out=None, where=True, **kwargs):
    """
    Compute the truth value of x1 OR x2 element-wise.

    Args:
        x1, x2 (numpoly.ndpoly):
            Logical OR is applied to the elements of `x1` and `x2`. If
            ``x1.shape != x2.shape``, they must be broadcastable to a common
            shape (which becomes the shape of the output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must have
            a shape that the inputs broadcast to. If not provided or `None`,
            a freshly-allocated array is returned. A tuple (possible only as a
            keyword argument) must have length equal to the number of outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value.
            Note that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        y (numpy.ndarray):
            Boolean result of the logical OR operation applied to the elements
            of `x1` and `x2`; the shape is determined by broadcasting.
            This is a scalar if both `x1` and `x2` are scalars.

    Examples:
        >>> numpoly.logical_or(True, False)
        True
        >>> numpoly.logical_or([True, False], [False, False])
        array([ True, False])
        >>> x = numpy.arange(5)
        >>> numpoly.logical_or(x < 1, x > 3)
        array([ True, False, False, False,  True])

    """
    x1 = numpoly.aspolynomial(x1)
    x2 = numpoly.aspolynomial(x2)
    coefficients1 = numpy.any(x1.coefficients, 0)
    coefficients2 = numpy.any(x2.coefficients, 0)
    return numpy.logical_or(
        coefficients1, coefficients2, out=out, where=where, **kwargs)



"""Find a polynomial with the given sequence of roots."""
import numpy
import numpoly


def polynomial_from_roots(
    seq_of_zeros,
    dtype=None,
):
    """
    Find the coefficients of a polynomial with the given sequence of roots.

    Returns the coefficients of the polynomial whose leading coefficient is one
    for the given sequence of zeros (multiple roots must be included in the
    sequence as many times as their multiplicity; see Examples). A square
    matrix (or array, which will be treated as a matrix) can also be given, in
    which case the coefficients of the characteristic polynomial of the matrix
    are returned.

    Args:
        seq_of_zeros (numpy.ndarray):
            A sequence of polynomial roots, or a square array or matrix object.
            Either shape (N,) or (N, N).
        dtype (Optional[numpy.dtype]):
            Any object that can be interpreted as a numpy data type.

    Returns:
        (numpoly.ndpoly):
            1-D polynomial which have `seq_of_zeros` as roots. Leading
            coefficient is always 1.

    Raises:
        ValueError:
            If input is the wrong shape (the input must be a 1-D or square
            2-D array).

    Examples:
        >>> numpoly.polynomial_from_roots((0, 0, 0))
        polynomial(q0**3)
        >>> numpoly.polynomial_from_roots((-0.5, 0, 0.5))
        polynomial(q0**3-0.25*q0)

    """
    exponent = numpy.arange(len(seq_of_zeros), -1, -1, dtype=int)
    basis = numpoly.variable(dtype=dtype)**exponent
    return numpoly.sum(numpy.poly(seq_of_zeros)*basis)



"""Return specified diagonals."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.diagonal)
def diagonal(a, offset=0, axis1=0, axis2=1):
    """
    Return specified diagonals.

    If `a` is 2-D, returns the diagonal of `a` with the given offset,
    i.e., the collection of elements of the form ``a[i, i+offset]``.  If
    `a` has more than two dimensions, then the axes specified by `axis1`
    and `axis2` are used to determine the 2-D sub-array whose diagonal is
    returned.  The shape of the resulting array can be determined by
    removing `axis1` and `axis2` and appending an index to the right equal
    to the size of the resulting diagonals.

    Args:
        a (numpoly.ndpoly):
            Array from which the diagonals are taken.
        offset (int):
            Offset of the diagonal from the main diagonal. Can be positive or
            negative. Defaults to main diagonal (0).
        axis1 (int):
            Axis to be used as the first axis of the 2-D sub-arrays from which
            the diagonals should be taken.  Defaults to first axis (0).
        axis2 (int):
            Axis to be used as the second axis of the 2-D sub-arrays from
            which the diagonals should be taken. Defaults to second axis (1).

    Returns:
        (numpoly.ndpoly):
            If `a` is 2-D, then a 1-D array containing the diagonal and of the
            same type as `a` is returned unless `a` is a `matrix`, in which
            case a 1-D array rather than a (2-D) `matrix` is returned in order
            to maintain backward compatibility.

            If ``a.ndim > 2``, then the dimensions specified by `axis1` and
            `axis2` are removed, and a new axis inserted at the end
            corresponding to the diagonal.

    Raises:
        ValueError:
            If the dimension of `a` is less than 2.

    Examples:
        >>> poly = numpoly.monomial(9).reshape(3, 3)
        >>> poly
        polynomial([[1, q0, q0**2],
                    [q0**3, q0**4, q0**5],
                    [q0**6, q0**7, q0**8]])
        >>> numpoly.diagonal(poly)
        polynomial([1, q0**4, q0**8])

    """
    a = numpoly.aspolynomial(a)
    out = numpy.diagonal(a.values, offset=offset, axis1=axis1, axis2=axis2)
    return numpoly.polynomial(out, names=a.names)



"""Evaluate polynomial."""
import logging
import numpy
import numpoly


def call(poly, args=(), kwargs=None):
    """
    Evaluate polynomial by inserting new values in to the indeterminants.

    Equivalent to calling the polynomial or using the ``__call__`` method.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to evaluate.
        args (Sequence[None, int, float, numpy.ndarray, numpoly.ndpoly]):
            Argument to evaluate indeterminants. Ordered positional by
            ``poly.indeterminants``. None values indicate that a variable is
            not to be evaluated, creating a partial evaluation.
        kwargs (Dict[str, Union[int, float, numpy.ndarray, numpoly.ndpoly]]):
            Same as ``args``, but positioned by name.

    Returns:
        (Union[numpy.ndarray, numpoly.ndpoly]):
            Evaluated polynomial. If the resulting array does not contain any
            indeterminants, an array is returned instead of a polynomial.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([[q0, q0-1], [q1, q1+q0]])
        >>> numpoly.call(poly)
        polynomial([[q0, q0-1],
                    [q1, q1+q0]])
        >>> poly
        polynomial([[q0, q0-1],
                    [q1, q1+q0]])
        >>> numpoly.call(poly, (1, 0))
        array([[1, 0],
               [0, 1]])
        >>> numpoly.call(poly, (1,), {"q1": [0, 1, 2]})
        array([[[1, 1, 1],
                [0, 0, 0]],
        <BLANKLINE>
               [[0, 1, 2],
                [1, 2, 3]]])
        >>> numpoly.call(poly, (q1,))
        polynomial([[q1, q1-1],
                    [q1, 2*q1]])
        >>> numpoly.call(poly, kwargs={"q1": q0-1, "q0": 2*q1})
        polynomial([[2*q1, 2*q1-1],
                    [q0-1, 2*q1+q0-1]])

    """
    logger = logging.getLogger(__name__)

    # Make sure kwargs contains all args and nothing but indeterminants:
    kwargs = kwargs if kwargs else {}
    args = list(args)
    for arg, indeterminant in zip(args, poly.names):
        if indeterminant in kwargs:
            raise TypeError(
                "multiple values for argument '%s'" % indeterminant)
        kwargs[indeterminant] = arg
    extra_args = [key for key in kwargs if key not in poly.names]
    if extra_args:
        raise TypeError("unexpected keyword argument '%s'" % extra_args[0])

    if not kwargs:
        return poly.copy()

    # Saturate kwargs with values not given:
    for name, indeterminant in zip(poly.names, poly.indeterminants):
        if kwargs.get(name, None) is None:
            kwargs[name] = indeterminant

    # There can only be one shape:
    ones = numpy.ones((), dtype=int)
    for value in kwargs.values():
        ones = ones * numpy.ones(numpoly.polynomial(value).shape, dtype=int)
    shape = poly.shape+ones.shape

    logger.debug("poly shape: %s", poly.shape)
    logger.debug("kwargs common shape: %s", ones.shape)
    logger.debug("output shape: %s", shape)

    # main loop:
    out = 0
    for exponent, coefficient in zip(poly.exponents, poly.coefficients):
        term = ones
        for power, name in zip(exponent, poly.names):
            term = term*kwargs[name]**power
        if isinstance(term, numpoly.ndpoly):
            tmp = numpoly.outer(coefficient, term)
        else:
            tmp = numpy.outer(coefficient, term)
        out = out+tmp.reshape(shape)

    if isinstance(out, numpoly.ndpoly):
        if out.isconstant():
            out = out.tonumpy()
        else:
            out, _ = numpoly.align_indeterminants(out, poly.indeterminants)
    return out



"""Convert the input to an polynomial array."""
from six import string_types

import numpy
import numpoly


def aspolynomial(
        poly_like=None,
        names=None,
        dtype=None,
):
    """
    Convert the input to an polynomial array.

    Args:
        poly_like (Any):
            Input to be converted to a `numpoly.ndpoly` polynomial type.
        names (str, Tuple[str, ...]):
            Name of the indeterminant variables. If possible to infer from
            ``poly_like``, this argument will be ignored.
        dtype (type, numpy.dtype):
            Data type used for the polynomial coefficients.

    Returns:
        (numpoly.ndpoly):
            Array interpretation of `poly_like`. No copy is performed if the
            input is already an ndpoly with matching indeterminants names and
            dtype.

    Examples:
        >>> q0 = numpoly.variable()
        >>> numpoly.polynomial(q0) is q0
        False
        >>> numpoly.aspolynomial(q0) is q0
        True

    """
    remain = False
    if isinstance(poly_like, numpoly.ndpoly):

        remain = (dtype is None or dtype == poly_like.dtype)
        if names is not None:
            if isinstance(names, numpoly.ndpoly):
                names = names.names
            if isinstance(names, string_types):
                names = [names]
            if len(names) == 1 and len(poly_like.names) > 1:
                names = ["{}{}".format(names[0], idx)
                         for idx in range(len(poly_like.indeterminants))]
            remain &= names == poly_like.names

    if remain:
        return poly_like
    return numpoly.polynomial(poly_like, names=names, dtype=dtype)



"""Inner product of two arrays."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.inner)
def inner(a, b):
    """
    Inner product of two arrays.

    Ordinary inner product of vectors for 1-D arrays (without complex
    conjugation), in higher dimensions a sum product over the last axes.

    """
    a, b = numpoly.align_exponents(a, b)
    return numpoly.sum(numpoly.multiply(a, b), axis=-1)



"""Return the element-wise square of the input."""
import numpy
import numpoly

from ..dispatch import implements
from .multiply import multiply


@implements(numpy.square)
def square(x, out=None, where=True, **kwargs):
    """
    Return the element-wise square of the input.

    Args:
        x (numpoly.ndpoly):
            Input data.
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        out (numpoly.ndpoly):
            Element-wise `x*x`, of the same shape and dtype as `x`.
            This is a scalar if `x` is a scalar.

    Examples:
        >>> numpoly.square([1j, 1])
        polynomial([(-1+0j), (1+0j)])
        >>> poly = numpoly.sum(numpoly.variable(2))
        >>> poly
        polynomial(q1+q0)
        >>> numpoly.square(poly)
        polynomial(q1**2+2*q0*q1+q0**2)

    """
    return multiply(x, x, out=out, where=where, **kwargs)



"""
Polynomial base class.

Under the hood, each of the polynomials are numpy structured arrays. The column
names are string representations corresponding to the polynomial exponents, and
the values are the coefficients. The indeterminant names are stored separately.
Numpoly is a wrapper on top of this.

From a development point of view, it is possible to view the underlying
structured array directly using the ``values`` attribute:

.. code:: python

    >>> q0, q1 = numpoly.variable(2)
    >>> poly = numpoly.polynomial(4*q0+3*q1-1)
    >>> array = poly.values
    >>> array  # doctest: +NORMALIZE_WHITESPACE
    array((-1, 4, 3),
          dtype=[(';;', '<i8'), ('<;', '<i8'), (';<', '<i8')])

Which, together with the indeterminant names, can be used to cast back the
array back to a polynomial:

.. code:: python

    >>> numpoly.aspolynomial(array, names=("q0", "q1"))
    polynomial(3*q1+4*q0-1)
"""
from __future__ import division
import logging
import re
from six import string_types

import numpy
from numpy.lib import recfunctions

from . import align, construct, dispatch, array_function, poly_function, option


REDUCE_MAPPINGS = {
    numpy.add: numpy.sum,
    numpy.multiply: numpy.prod,
    numpy.logical_and: numpy.all,
    numpy.logical_or: numpy.any,
    numpy.maximum: numpy.amax,
    numpy.minimum: numpy.amin,
}
ACCUMULATE_MAPPINGS = {
    numpy.add: numpy.cumsum,
    numpy.multiply: numpy.cumprod,
}


class FeatureNotSupported(ValueError):
    """Error for features in numpy not supported in Numpoly."""


class ndpoly(numpy.ndarray):  # pylint: disable=invalid-name
    """
    Polynomial as numpy array.

    An array object represents a multidimensional, homogeneous polynomial array
    of fixed-size items. An associated data-type object describes the format of
    each element in the array (its byte-order, how many bytes it occupies in
    memory, whether it is an integer, a floating point number, or something
    else, etc.)

    Though possible, it is not recommended to construct polynomials using
    ``ndpoly`` for basic polynomial array construction. Instead the user should
    be consider using construction functions like `variable`, `monomial`,
    `polynomial`, etc.

    Examples:
        >>> poly = ndpoly(
        ...     exponents=[(0, 1), (0, 0)], shape=(3,))
        >>> poly.values[";<"] = 4, 5, 6
        >>> poly.values[";;"] = 1, 2, 3
        >>> numpy.array(poly.coefficients)
        array([[4, 5, 6],
               [1, 2, 3]])
        >>> poly
        polynomial([4*q1+1, 5*q1+2, 6*q1+3])
        >>> poly[0]
        polynomial(4*q1+1)

    """

    # =================================================
    # Stuff to get subclassing of ndarray to run smooth
    # =================================================

    __array_priority__ = 16

    _dtype = None
    """
    Underlying structure array's actual dtype.

    Column names correspond to polynomial exponents.
    The numerical values can be calculated using the formula:
    ``poly._dtype.view(numpy.uint32)-poly.KEY_OFFSET``.
    """

    keys = None
    """
    The raw names of the coefficients.

    One-to-one with `exponents`, but as string as to be compatible with numpy
    structured array. Unlike the exponents, that are useful for mathematical
    manipulation, the keys are useful as coefficient lookup.

    The column names in the underlying structured array dtype
    ``ndpoly._dtype``.
    """

    names = None
    """
    Same as `indeterminants`, but only the names as string.

    Positional list of indeterminant names.
    """

    allocation = None
    """
    The number of polynomial coefficients allocated to polynomial.
    """

    # Numpy structured array names doesn't like characters reserved by Python.
    # The largest index found with this property is 58: ':'.
    # Above this, everything looks like it works as expected.
    KEY_OFFSET = 59
    """
    Internal number off-set between exponent and its stored value.

    Exponents are stored in structured array names, which are limited to not accept the letter ':'. By adding an offset between the represented value and the stored value, the letter ':' is skipped.
    """

    def __new__(
            cls,
            exponents=((0,),),
            shape=(),
            names=None,
            dtype=None,
            allocation=None,
            **kwargs
    ):
        """
        Class constructor.

        Args:
            exponents (numpy.ndarray):
                The exponents in an integer array with shape ``(N, D)``, where
                ``N`` is the number of terms in the polynomial sum and ``D`` is
                the number of dimensions.
            shape (Tuple[int, ...]):
                Shape of created array.
            names (Union[None, str, Tuple[str], numpoly.ndpoly]):
                The name of the indeterminant variables in the polynomial. If
                polynomial, inherent from it. Else, pass argument to
                `numpoly.symbols` to create the indeterminants names. If only
                one name is provided, but more than one is required,
                indeterminants will be extended with an integer index. If
                omitted, use ``numpoly.get_options()["default_varname"]``.
            dtype (Optional[numpy.dtype]):
                Any object that can be interpreted as a numpy data type.
            allocation (Optional[int]):
                The maximum number of polynomial exponents. If omitted, use
                length of exponents for allocation.
            kwargs:
                Extra arguments passed to `numpy.ndarray` constructor.

        """
        exponents = numpy.array(exponents, dtype=numpy.uint32)

        if numpy.prod(exponents.shape):
            keys = (exponents+cls.KEY_OFFSET).flatten()
            keys = keys.view("U%d" % exponents.shape[-1])
            keys = numpy.array(keys, dtype="U%d" % (exponents.shape[-1]))
        else:
            keys = numpy.zeros(0, dtype="U1")

        if allocation is None:
            allocation = 2*len(keys)
        assert isinstance(allocation, int) and allocation >= len(keys), (
            "Not enough memory allocated; increase 'allocation'")
        if allocation > len(keys):
            allocation_ = numpy.arange(allocation-len(keys), len(keys))
            allocation_ = [str(s) for s in allocation_]
            keys = numpy.concatenate([keys, allocation_])

        if names is None:
            names = option.get_options()["default_varname"]
            names = construct.symbols("%s:%d" % (names, exponents.shape[-1])).names
        elif isinstance(names, string_types):
            names = construct.symbols(names).names
        elif isinstance(names, ndpoly):
            names = names.names
        for name in names:
            assert re.search(option.get_options()["varname_filter"], name), (
                "invalid polynomial name; "
                "expected format: %r" % option.get_options()["varname_filter"])

        dtype = int if dtype is None else dtype
        dtype_ = numpy.dtype([(key, dtype) for key in keys])

        obj = super(ndpoly, cls).__new__(
            cls, shape=shape, dtype=dtype_, **kwargs)
        obj._dtype = numpy.dtype(dtype)  # pylint: disable=protected-access
        obj.keys = keys
        obj.names = tuple(names)
        obj.allocation = allocation
        return obj

    def __array_finalize__(self, obj):
        """Finalize numpy constructor."""
        if obj is None:
            return
        self.keys = getattr(obj, "keys", None)
        self.names = getattr(obj, "names", None)
        self.allocation = getattr(obj, "allocation", None)
        self._dtype = getattr(obj, "_dtype", None)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Dispatch method for operators."""
        if method == "reduce":
            ufunc = REDUCE_MAPPINGS[ufunc]
        elif method == "accumulate":
            ufunc = ACCUMULATE_MAPPINGS[ufunc]
        elif method != "__call__":
            raise FeatureNotSupported("Method '%s' not supported." % method)
        if ufunc not in dispatch.UFUNC_COLLECTION:
            raise FeatureNotSupported("ufunc '%s' not supported." % ufunc)
        return dispatch.UFUNC_COLLECTION[ufunc](*inputs, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        """Dispatch method for functions."""
        logger = logging.getLogger(__name__)
        fname = func.__name__
        if func not in dispatch.FUNCTION_COLLECTION:
            raise FeatureNotSupported(
                "function '%s' not supported by numpoly." % fname)

        # notify that numpy.save* works, but numpy.load* fails
        if fname in ("save", "savez", "savez_compressed"):
            logger.warning("""\
numpy.%s used to store numpoly.ndpoly (instead of numpoly.%s).
This works, but restoring requires using numpoly.load, \
as numpy.load will not work as expected.""" % (fname, fname))
        elif fname == "savetxt":
            logger.warning("""\
numpy.%s used to store numpoly.ndpoly (instead of numpoly.%s).
This works, but restoring requires using numpoly.loadtxt, \
as numpy.loadtxt will not work as expected.""" % (fname, fname))
        return dispatch.FUNCTION_COLLECTION[func](*args, **kwargs)

    # ======================================
    # Properties specific for ndpoly objects
    # ======================================

    @property
    def coefficients(self):
        """
        Polynomial coefficients.

        Together with exponents defines the polynomial form.

        Examples:
            >>> q0, q1 = numpoly.variable(2)
            >>> poly = numpoly.polynomial([2*q0**4, -3*q1**2+14])
            >>> poly
            polynomial([2*q0**4, -3*q1**2+14])
            >>> numpy.array(poly.coefficients)
            array([[ 0, 14],
                   [ 0, -3],
                   [ 2,  0]])

        """
        if not self.size:
            return []
        out = numpy.empty((len(self.keys),) + self.shape, dtype=self._dtype)
        for idx, key in enumerate(self.keys):
            out[idx] = numpy.ndarray.__getitem__(self, key)
        return list(out)

    @property
    def exponents(self):
        """
        Polynomial exponents.

        2-dimensional where the first axis is the same length as coefficients
        and the second is the length of the indeterminant names.

        Examples:
            >>> q0, q1 = numpoly.variable(2)
            >>> poly = numpoly.polynomial([2*q0**4, -3*q1**2+14])
            >>> poly
            polynomial([2*q0**4, -3*q1**2+14])
            >>> poly.exponents
            array([[0, 0],
                   [0, 2],
                   [4, 0]], dtype=uint32)

        """
        exponents = self.keys.astype("U%d" % len(self.names))
        exponents = exponents.view(numpy.uint32)-self.KEY_OFFSET
        if numpy.prod(exponents.shape):
            exponents = exponents.reshape(len(self.keys), -1)
        assert len(exponents) >= 0
        assert len(exponents.shape) == 2
        return exponents

    @staticmethod
    def from_attributes(
            exponents,
            coefficients,
            names=None,
            dtype=None,
            allocation=None,
            retain_coefficients=None,
            retain_names=None,
    ):
        """
        Construct polynomial from polynomial attributes.

        Args:
            exponents (numpy.ndarray):
                The exponents in an integer array with shape ``(N, D)``, where
                ``N`` is the number of terms in the polynomial sum and ``D`` is
                the number of dimensions.
            coefficients (Iterable[numpy.ndarray]):
                The polynomial coefficients. Must correspond to `exponents` by
                having the same length ``N``.
            names (Union[None, str, Tuple[str, ...], numpoly.ndpoly]):
                The indeterminants names, either as string names or as
                simple polynomials. Must correspond to the exponents by having
                length 1 or ``D``. If omitted, use
                ``numpoly.get_options()["default_varname"]``.
            dtype (Optional[numpy.dtype]):
                The data type of the polynomial. If omitted, extract from
                `coefficients`.
            allocation (Optional[int]):
                The maximum number of polynomial exponents. If omitted, use
                length of exponents for allocation.
            retain_coefficients (Optional[bool]):
                Do not remove redundant coefficients. If omitted use global
                defaults.
            retain_names (Optional[bool]):
                Do not remove redundant names. If omitted use global defaults.

        Returns:
            (numpoly.ndpoly):
                Polynomials with attributes defined by input.

        Examples:
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[0]], coefficients=[4])
            polynomial(4)
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[1]], coefficients=[[1, 2, 3]])
            polynomial([q0, 2*q0, 3*q0])
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[0], [1]], coefficients=[[0, 1], [1, 1]])
            polynomial([q0, q0+1])
            >>> numpoly.ndpoly.from_attributes(
            ...     exponents=[[0, 1], [1, 1]], coefficients=[[0, 1], [1, 1]])
            polynomial([q0*q1, q0*q1+q1])

        """
        return construct.polynomial_from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            names=names,
            dtype=dtype,
            allocation=allocation,
            retain_coefficients=retain_coefficients,
            retain_names=retain_names,
        )

    @property
    def indeterminants(self):
        """
        Polynomial indeterminants.

        Secondary polynomial only consisting of an array of simple independent
        variables found in the polynomial array.

        Examples:
            >>> q0, q1 = numpoly.variable(2)
            >>> poly = numpoly.polynomial([2*q0**4, -3*q1**2+14])
            >>> poly
            polynomial([2*q0**4, -3*q1**2+14])
            >>> poly.indeterminants
            polynomial([q0, q1])

        """
        return construct.polynomial_from_attributes(
            exponents=numpy.eye(len(self.names), dtype=int),
            coefficients=numpy.eye(len(self.names), dtype=int),
            names=self.names,
        )

    @property
    def values(self):
        """
        Expose the underlying structured array.

        Typically used for operator dispatching and not for use to conversion.

        Examples:
            >>> numpoly.variable(1).values
            array((1,), dtype=[('<', '<i8')])
            >>> numpoly.variable(2).values
            array([(1, 0), (0, 1)], dtype=[('<;', '<i8'), (';<', '<i8')])

        """
        return numpy.ndarray(
            shape=self.shape,
            dtype=[(key, self.dtype) for key in self.keys],
            buffer=self.data
        )

    def isconstant(self):
        """
        Check if a polynomial is constant or not.

        Returns:
            (bool):
                True if all elements in array are constant.

        Examples:
            >>> q0 = numpoly.variable()
            >>> q0.isconstant()
            False
            >>> numpoly.polynomial([1, 2]).isconstant()
            True
            >>> numpoly.polynomial([1, q0]).isconstant()
            False

        """
        return poly_function.isconstant(self)

    def todict(self):
        """
        Cast to dict where keys are exponents and values are coefficients.

        Returns:
            (Dict[Tuple[int, ...], numpy.ndarray]):
                Dictionary where keys are exponents and values are
                coefficients.

        Examples:
            >>> q0, q1 = numpoly.variable(2)
            >>> poly = 2*q0**4-3*q1**2+14
            >>> poly
            polynomial(2*q0**4-3*q1**2+14)
            >>> poly.todict() == {(0, 0): 14, (4, 0): 2, (0, 2): -3}
            True

        """
        return {tuple(exponent): coefficient
                for exponent, coefficient in zip(
                    self.exponents, self.coefficients)}

    def tonumpy(self):
        """
        Cast polynomial to numpy.ndarray, if possible.

        Returns:
            (numpy.ndarray):
                Same as object, but cast to `numpy.ndarray`.

        Raises:
            numpoly.baseclass.FeatureNotSupported:
                When polynomial include indeterminats, casting to numpy.

        Examples:
            >>> numpoly.polynomial([1, 2]).tonumpy()
            array([1, 2])

        """
        return poly_function.tonumpy(self)

    # =============================================
    # Override numpy properties to work with ndpoly
    # =============================================

    @property
    def dtype(self):
        """Datatype of the polynomial coefficients."""
        return self._dtype

    def astype(self, dtype, **kwargs):
        """Wrap ndarray.astype."""
        coefficients = [coefficient.astype(dtype, **kwargs)
                        for coefficient in self.coefficients]
        return construct.polynomial_from_attributes(
            exponents=self.exponents,
            coefficients=coefficients,
            names=self.names,
            allocation=self.allocation,
            dtype=dtype,
        )

    def diagonal(self, offset=0, axis1=0, axis2=1):
        """Wrap ndarray.diagonal."""
        return array_function.diagonal(
            self, offset=offset, axis1=axis1, axis2=axis2)

    def round(self, decimals=0, out=None):
        """Wrap ndarray.round."""
        # Not sure why it is required. Likely a numpy bug.
        return array_function.around(self, decimals=decimals, out=out)

    def max(self, axis=None, out=None, keepdims=False, **kwargs):
        """Wrap ndarray.max."""
        return array_function.max(self, axis=axis, out=out,
                                  keepdims=keepdims, **kwargs)

    def min(self, axis=None, out=None, keepdims=False, **kwargs):
        """Wrap ndarray.min."""
        return array_function.min(self, axis=axis, out=out,
                                  keepdims=keepdims, **kwargs)

    # ============================================================
    # Override dunder methods that isn't dealt with by dispatching
    # ============================================================

    def __call__(self, *args, **kwargs):
        """
        Evaluate polynomial by inserting new values in to the indeterminants.

        Args:
            args (int, float, numpy.ndarray, numpoly.ndpoly):
                Argument to evaluate indeterminants. Ordered positional by
                ``self.indeterminants``.
            kwargs (int, float, numpy.ndarray, numpoly.ndpoly):
                Same as ``args``, but positioned by name.

        Returns:
            (Union[numpy.ndarray, numpoly.ndpoly]):
                Evaluated polynomial. If the resulting array does not contain
                any indeterminants, an array is returned instead of a
                polynomial.

        Examples:
            >>> q0, q1 = numpoly.variable(2)
            >>> poly = numpoly.polynomial(
            ...     [[q0, q0-1], [q1, q1+q0]])
            >>> poly(1, q1=[0, 1, 2])
            array([[[1, 1, 1],
                    [0, 0, 0]],
            <BLANKLINE>
                   [[0, 1, 2],
                    [1, 2, 3]]])

        """
        return poly_function.call(self, args, kwargs)

    def __eq__(self, other):
        """Left equality."""
        return array_function.equal(self, other)

    def __getitem__(self, index):
        """
        Get array item or slice.

        Args:
            index (Union[int, str, Tuple[int, ...], numpy.ndarray]):
                The index to extract. If string type, extract using th
                underlying structured array data type. Else, extract as
                numpy.ndarray.

        Examples:
            >>> q0, q1 = numpoly.variable(2)
            >>> poly = numpoly.polynomial([[1-4*q0, q0**2], [q1-3, q0*q1*q1]])
            >>> poly
            polynomial([[-4*q0+1, q0**2],
                        [q1-3, q0*q1**2]])
            >>> poly[0]
            polynomial([-4*q0+1, q0**2])
            >>> poly[:, 1]
            polynomial([q0**2, q0*q1**2])
            >>> poly["<;"]
            array([[-4,  0],
                   [ 0,  0]])

        """
        if isinstance(index, string_types):
            return numpy.asarray(super(ndpoly, self).__getitem__(index))
        return construct.polynomial_from_attributes(
            exponents=self.exponents,
            coefficients=[coeff[index] for coeff in self.coefficients],
            names=self.names,
        )

    # def __setitem__(self, index, other):
    #     other = construct.aspolynomial(other)
    #     difference =  set(other.names).difference(self.names)
    #     assert not difference, (
    #         "polynomial does not contain indeterminants: %s" % difference)
    #     assert self.ndim >= other.ndim
    #     other, _ = align.align_polynomials(other, self)
    #     for key in other.keys:
    #         if key not in self.keys:
    #             for idx, key_ in enumerate(self.values.dtype.names):
    #                 if key_.isdigit():
    #                     break
    #             else:
    #                 fail
    #             names = list(self.values.dtype.names)
    #             names[idx] = key
    #         self.values[key][index] = other.values[key][index]


    def __iter__(self):
        """Iterate polynomial array."""
        coefficients = numpy.array(list(self.coefficients))
        return iter(construct.polynomial_from_attributes(
            exponents=self.exponents,
            coefficients=coefficients[:, idx],
            names=self.names,
        ) for idx in range(len(self)))

    def __ne__(self, other):
        """Not equal."""
        return array_function.not_equal(self, other)

    def __repr__(self):
        """Canonical string representation."""
        return array_function.array_repr(self)

    def __str__(self):
        """Pretty string representation."""
        return array_function.array_str(self)

    def __truediv__(self, value):
        """Return self/value."""
        return poly_function.poly_divide(self, value)

    def __rtruediv__(self, value):
        """Return value/self."""
        return poly_function.poly_divide(value, self)

    def __div__(self, value):
        """Return self/value."""
        return poly_function.poly_divide(self, value)

    def __rdiv__(self, value):
        """Return value/self."""
        return poly_function.poly_divide(value, self)

    def __mod__(self, value):
        """Return self%value."""
        return poly_function.poly_remainder(self, value)

    def __rmod__(self, value):
        """Return value%self."""
        return poly_function.poly_remainder(value, self)

    def __divmod__(self, value):
        """Return divmod(self, value)."""
        return poly_function.poly_divmod(self, value)

    def __rdivmod__(self, value):
        """Return divmod(value, self)."""
        return poly_function.poly_divmod(value, self)

    def __reduce__(self):
        """Extract state to be pickled."""
        return (construct.polynomial_from_attributes,
                (self.exponents, self.coefficients, self.names,
                 self.dtype, self.allocation, False))



"""Expand the shape of an array."""
import numpy
import numpoly

from ..dispatch import implements

@implements(numpy.expand_dims)
def expand_dims(a, axis):
    """
    Expand the shape of an array.

    Insert a new axis that will appear at the `axis` position in the expanded
    array shape.

    Args:
        a (numpoly.ndarray):
            Input array.
        axis (int):
            Position in the expanded axes where the new axis is placed.

    Returns:
        (ndpoly):
            View of `a` with the number of dimensions increased by one.

    Examples:
        >>> poly = numpoly.variable(2)
        >>> numpoly.expand_dims(poly, axis=0)
        polynomial([[q0, q1]])
        >>> numpoly.expand_dims(poly, axis=1)
        polynomial([[q0],
                    [q1]])

    """
    a = numpoly.aspolynomial(a)
    out = numpy.expand_dims(a.values, axis=axis)
    return numpoly.polynomial(out, names=a.indeterminants)



"""Functions for constructing polynomials."""
from .polynomial import polynomial
from .aspolynomial import aspolynomial
from .clean import (
    clean_attributes, remove_redundant_coefficients, remove_redundant_names)
from .from_attributes import polynomial_from_attributes
from .from_roots import polynomial_from_roots
from .monomial import monomial
from .symbols import symbols
from .variable import variable



"""Convert numpoly.ndpoly to sympy polynomial."""
import numpy


def to_sympy(poly):
    """
    Convert numpoly object to sympy object, or array of sympy objects.

    Args:
        poly (numpoly.ndpoly):
            Polynomial object to convert to sympy.

    Returns:
        (numpy.ndarray, sympy.core.expr.Expr):
            If scalar, a sympy expression object, or if array, numpy.array with
            expression object values.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([[1, q0**3], [q1-1, -3*q0]])
        >>> sympy_poly = to_sympy(poly)
        >>> sympy_poly
        array([[1, q0**3],
               [q1 - 1, -3*q0]], dtype=object)
        >>> type(sympy_poly.item(-1))
        <class 'sympy.core.mul.Mul'>

    """
    if poly.shape:
        return numpy.array([to_sympy(poly_) for poly_ in poly])
    from sympy import symbols
    locals_ = dict(zip(poly.names, symbols(poly.names)))
    polynomial = eval(str(poly), locals_, {})  # pylint: disable=eval-used
    return polynomial



"""Repeat elements of an array."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.repeat)
def repeat(a, repeats, axis=0):
    """
    Repeat elements of an array.

    Args:
        a (numpoly.ndpoly):
            Input array.
        repeats (Union[int, numpy.ndarray]):
            The number of repetitions for each element. `repeats` is
            broadcasted to fit the shape of the given axis.
        axis (Optional[int]):
            The axis along which to repeat values. By default, use the
            flattened input array, and return a flat output array.

    Returns:
        (ndarray):
            Output array which has the same shape as `a`, except along the
            given axis.

    Examples:
        >>> q0 = numpoly.variable()
        >>> numpoly.repeat(q0, 4)
        polynomial([q0, q0, q0, q0])
        >>> poly = numpoly.polynomial([[1, q0-1], [q0**2, q0]])
        >>> numpoly.repeat(poly, 2)
        polynomial([[1, q0-1],
                    [1, q0-1],
                    [q0**2, q0],
                    [q0**2, q0]])
        >>> numpoly.repeat(poly, 3, axis=1)
        polynomial([[1, 1, 1, q0-1, q0-1, q0-1],
                    [q0**2, q0**2, q0**2, q0, q0, q0]])
        >>> numpoly.repeat(poly, [1, 2], axis=0)
        polynomial([[1, q0-1],
                    [q0**2, q0],
                    [q0**2, q0]])

    """
    a = numpoly.aspolynomial(a)
    result = numpy.repeat(a.values, repeats=repeats, axis=axis)
    return numpoly.aspolynomial(result, names=a.indeterminants)



"""Simple constructor to create single variables to create polynomials."""
import numpoly


def variable(dimensions=1, asarray=False, dtype="i8", allocation=None):
    """
    Construct variables that can be used to construct polynomials.

    Args:
        dimensions (int):
            Number of dimensions in the array.
        asarray (bool):
            Enforce output as array even in the case where there is only one
            variable.
        dtype (numpy.dtype):
            The data type of the polynomial coefficients.
        allocation (Optional[int]):
            The maximum number of polynomial exponents. If omitted, use
            length of exponents for allocation.

    Returns:
        (chaospy.poly.polynomial):
            Polynomial array with unit components in each dimension.

    Examples:
        >>> numpoly.variable()
        polynomial(q0)
        >>> q0, q1, q2 = numpoly.variable(3)
        >>> q1+1
        polynomial(q1+1)
        >>> numpoly.polynomial([q2**3, q1+q2, 1])
        polynomial([q2**3, q2+q1, 1])

    """
    return numpoly.symbols(
        names="%s:%d" % (numpoly.get_options()["default_varname"], dimensions),
        asarray=asarray,
        dtype=dtype,
        allocation=allocation,
    )



"""Construct an array from an index array and a set of arrays to choose from."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.choose)
def choose(a, choices, out=None, mode="raise"):
    """
    Construct an array from an index array and a set of arrays to choose from.

    First of all, if confused or uncertain, definitely look at the Examples -
    in its full generality, this function is less simple than it might
    seem from the following code description (below ndi =
    `numpy.lib.index_tricks`):

    ``np.choose(a,c) == np.array([c[a[I]][I] for I in ndi.ndindex(a.shape)])``.

    But this omits some subtleties. Here is a fully general summary:

    Given an "index" array (`a`) of integers and a sequence of `n` arrays
    (`choices`), `a` and each choice array are first broadcast, as necessary,
    to arrays of a common shape; calling these *Ba* and *Bchoices[i], i =
    0,...,n-1* we have that, necessarily, ``Ba.shape == Bchoices[i].shape``
    for each `i`.  Then, a new array with shape ``Ba.shape`` is created as
    follows:

    * if ``mode=raise`` (the default), then, first of all, each element of
      `a` (and thus `Ba`) must be in the range `[0, n-1]`; now, suppose that
      `i` (in that range) is the value at the `(j0, j1, ..., jm)` position
      in `Ba` - then the value at the same position in the new array is the
      value in `Bchoices[i]` at that same position;

    * if ``mode=wrap``, values in `a` (and thus `Ba`) may be any (signed)
      integer; modular arithmetic is used to map integers outside the range
      `[0, n-1]` back into that range; and then the new array is constructed
      as above;

    * if ``mode=clip``, values in `a` (and thus `Ba`) may be any (signed)
      integer; negative integers are mapped to 0; values greater than `n-1`
      are mapped to `n-1`; and then the new array is constructed as above.

    Args:
        a (int, numpy.ndarray):
            This array must contain integers in `[0, n-1]`, where `n` is the
            number of choices, unless ``mode=wrap`` or ``mode=clip``, in which
            cases any integers are permissible.
        choices (Sequence[numpoly.ndpoly]):
            Choice arrays. `a` and all of the choices must be broadcastable to
            the same shape.  If `choices` is itself an array (not recommended),
            then its outermost dimension (i.e., the one corresponding to
            ``choices.shape[0]``) is taken as defining the "sequence".
        out (Optional[numpoly.ndpoly]):
            If provided, the result will be inserted into this array. It should
            be of the appropriate shape and dtype. Note that `out` is always
            buffered if `mode='raise'`; use other modes for better performance.
        mode (Optional[str]):
            {'raise' (default), 'wrap', 'clip'}, optional
            Specifies how indices outside `[0, n-1]` will be treated:

              * 'raise' : an exception is raised
              * 'wrap' : value becomes value mod `n`
              * 'clip' : values < 0 are mapped to 0, values > n-1 are mapped to n-1

    Returns:
        merged_array (numpoly.ndpoly):
            The merged result.

    Raises:
        ValueError: shape mismatch
            If `a` and each choice array are not all broadcastable to the same
            shape.

    Notes:
        To reduce the chance of misinterpretation, even though the following
        "abuse" is nominally supported, `choices` should neither be, nor be
        thought of as, a single array, i.e., the outermost sequence-like
        container should be either a list or a tuple.

    Examples:
        >>> poly1 = numpoly.monomial(3, dimensions="q0")
        >>> poly2 = numpoly.monomial(3, dimensions="q1")
        >>> choices = numpoly.outer(poly1, poly2)
        >>> choices
        polynomial([[1, q1, q1**2],
                    [q0, q0*q1, q0*q1**2],
                    [q0**2, q0**2*q1, q0**2*q1**2]])
        >>> numpoly.choose([1, 2, 0], choices)
        polynomial([q0, q0**2*q1, q1**2])
        >>> numpoly.choose([1, 3, 0], choices, mode="wrap")
        polynomial([q0, q1, q1**2])

    """
    choices = numpoly.aspolynomial(choices)
    result = numpy.choose(a, choices=choices.values, out=out, mode=mode)
    return numpoly.aspolynomial(result, names=choices.indeterminants)



"""Split an array into multiple sub-arrays vertically (row-wise)."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.vsplit)
def vsplit(ary, indices_or_sections):
    """
    Split an array into multiple sub-arrays vertically (row-wise).

    Please refer to the ``split`` documentation.  ``vsplit`` is equivalent
    to ``split`` with `axis=0` (default), the array is always split along the
    first axis regardless of the array dimension.

    See Also:
        split : Split an array into multiple sub-arrays of equal size.

    Examples:
        >>> poly = numpoly.monomial(8).reshape(4, 2)
        >>> poly
        polynomial([[1, q0],
                    [q0**2, q0**3],
                    [q0**4, q0**5],
                    [q0**6, q0**7]])
        >>> part1, part2 = numpoly.vsplit(poly, 2)
        >>> part1
        polynomial([[1, q0],
                    [q0**2, q0**3]])
        >>> part2
        polynomial([[q0**4, q0**5],
                    [q0**6, q0**7]])
        >>> part1, part2, part3 = numpoly.vsplit(poly, [1, 2])
        >>> part1
        polynomial([[1, q0]])
        >>> part3
        polynomial([[q0**4, q0**5],
                    [q0**6, q0**7]])

    """
    ary = numpoly.aspolynomial(ary)
    results = numpy.vsplit(ary.values, indices_or_sections=indices_or_sections)
    return [
        numpoly.polynomial(
            result, names=ary.indeterminants, allocation=ary.allocation)
        for result in results
    ]



"""Return element-wise quotient and remainder simultaneously."""
import numpy
import numpoly

from ..dispatch import implements_ufunc

DIVMOD_ERROR_MSG = """
Division-remainder involving polynomial division differs from numerical division;
Use ``numpoly.poly_divmod`` to get polynomial division-remainder."""


@implements_ufunc(numpy.divmod)
def divmod(x1, x2, out=(None, None), where=True, **kwargs):
    """
    Return element-wise quotient and remainder simultaneously.

    ``numpoly.divmod(x, y)`` is equivalent to ``(x // y, x % y)``, but faster
    because it avoids redundant work. It is used to implement the Python
    built-in function ``divmod`` on arrays.

    Args:
        x1 (numpoly.ndpoly):
            Dividend array.
        x2 (numpoly.ndpoly):
            Divisor array. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Tuple[Optional[numpoly.ndpoly], Optional[numpoly.ndpoly]]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (Tuple[numpoly.ndpoly, numpoly.ndpoly]):
            Element-wise quotient and remainder resulting from floor division.

    Raises:
        numpoly.baseclass.FeatureNotSupported:
            If either `x1` or `x2` contains indeterminants, numerical division
            is no longer possible and an error is raised instead. For
            polynomial division-remainder see ``numpoly.poly_divmod``.

    Examples:
        >>> numpoly.divmod([1, 22, 444], 4)
        (polynomial([0, 5, 111]), polynomial([1, 2, 0]))

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    if not x1.isconstant() or not x2.isconstant():
        raise numpoly.FeatureNotSupported(DIVMOD_ERROR_MSG)
    quotient, remainder = numpy.divmod(
        x1.tonumpy(), x2.tonumpy(), out=out, where=where, **kwargs)
    return numpoly.polynomial(quotient), numpoly.polynomial(remainder)



"""Implementation wrapper."""
import numpy
import numpoly

FUNCTION_COLLECTION = {}
UFUNC_COLLECTION = {}


def implements_function(*array_functions):
    """Register __array_function__."""
    def decorator(numpoly_function):
        """Register function."""
        for func in array_functions:
            assert func not in FUNCTION_COLLECTION, "%s already implemented" % func
            FUNCTION_COLLECTION[func] = numpoly_function
        return numpoly_function
    return decorator


def implements_ufunc(*array_methods):
    """Register __array_ufunc__."""
    def decorator(numpoly_function):
        """Register function."""
        for func in array_methods:
            assert func not in UFUNC_COLLECTION, "%s already implemented" % func
            UFUNC_COLLECTION[func] = numpoly_function
        return numpoly_function
    return decorator


def implements(*array_functions):
    """Register __array_function__ and __array_ufunc__."""
    def decorator(numpoly_function):
        """Register function."""
        for func in array_functions:
            assert func not in FUNCTION_COLLECTION, "%s already implemented" % func
            FUNCTION_COLLECTION[func] = numpoly_function
            assert func not in UFUNC_COLLECTION, "%s already implemented" % func
            UFUNC_COLLECTION[func] = numpoly_function
        return numpoly_function

    return decorator


def simple_dispatch(
        numpy_func,
        inputs,
        out=None,
        **kwargs
):
    """
    Dispatch function between numpy and numpoly.

    Assumes that evaluation can be performed on the coefficients alone and that
    there are no change to the polynomials.

    Args:
        numpy_func (Callable):
            The numpy function to evaluate `inputs` on.
        inputs (Iterable[numpoly.ndpoly]):
            One or more input arrays.
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        kwargs:
            Keyword args passed to `numpy_func`.

    Returns:
        (numpoly.ndpoly):
            Polynomial, where the coefficients from `input` are passed to
            `numpy_func` to create the output coefficients.

    """
    inputs = numpoly.align_polynomials(*inputs)
    no_output = out is None
    keys = inputs[0].keys if no_output else out.keys
    for key in keys:

        if out is None:
            tmp = numpy_func(*[poly[key] for poly in inputs], **kwargs)
            out = numpoly.ndpoly(
                exponents=inputs[0].exponents,
                shape=tmp.shape,
                names=inputs[0].indeterminants,
                dtype=tmp.dtype,
            )
            out[key] = tmp

        elif no_output:
            out[key] = numpy_func(*[poly[key] for poly in inputs], **kwargs)

        else:
            tmp = numpy_func(
                *[poly[key] for poly in inputs], out=out[key], **kwargs)

    if no_output:
        out = numpoly.clean_attributes(out)
    return out



"""Clean up polynomial attributes."""
from six import string_types
import numpy
import numpoly


class PolynomialConstructionError(ValueError):
    """Error related to construction of polynomial."""


def clean_attributes(
        poly,
        retain_coefficients=None,
        retain_names=None,
):
    """
    Clean up polynomial attributes.

    Some operations results in polynomial with structures that are redundant.
    This includes extra unused indeterminants, and extra terms consisting of
    only zeros.

    Args:
        poly (numpoly.ndpoly):
            Polynomial to clean up.
        retain_coefficients (Optional[bool]):
            Do not remove redundant coefficients. If omitted use global
            defaults.
        retain_names (Optional[bool]):
            Do not remove redundant names. If omitted use global defaults.

    Returns:
        Same as `poly`, but with attributes cleaned up.

    Examples:
        >>> q0, _ = numpoly.align_polynomials(*numpoly.variable(2))
        >>> q0.indeterminants
        polynomial([q0, q1])
        >>> q0.exponents
        array([[1, 0],
               [0, 1]], dtype=uint32)
        >>> q0 = numpoly.clean_attributes(
        ...     q0, retain_coefficients=False, retain_names=False)
        >>> q0.indeterminants
        polynomial([q0])
        >>> q0.exponents
        array([[1]], dtype=uint32)

    """
    return numpoly.ndpoly.from_attributes(
        exponents=poly.exponents,
        coefficients=poly.coefficients,
        names=poly.names,
        dtype=poly.dtype,
        retain_coefficients=retain_coefficients,
        retain_names=retain_names,
    )


def postprocess_attributes(
        exponents,
        coefficients,
        names=None,
        retain_coefficients=None,
        retain_names=None,
):
    """
    Clean up polynomial attributes.

    Args:
        exponents (numpy.ndarray):
            The exponents in an integer array with shape ``(N, D)``, where
            ``N`` is the number of terms in the polynomial sum and ``D`` is
            the number of dimensions.
        coefficients (Sequence[numpy.ndarray]):
            The polynomial coefficients. Must correspond to `exponents` by
            having the same length ``N``.
        names (Union[None, Sequence[str], numpoly.ndpoly]):
            The indeterminant names, either as string names or as
            simple polynomials. Must correspond to the exponents by having
            length ``D``.
        retain_coefficients (Optional[bool]):
            Do not remove redundant coefficients. If omitted use global
            defaults.
        retain_names (Optional[bool]):
            Do not remove redundant names. If omitted use global defaults.

    Returns:
        (numpoly.ndarray, List[numpy.ndarray], Optional[Tuple[str, ...]]):
            Same as input, but post-processed.

    """
    exponents = numpy.asarray(exponents)
    if exponents.ndim != 2:
        raise PolynomialConstructionError(
            "expected exponents.ndim == 2; found %d" % exponents.ndim)

    coefficients = [numpy.asarray(coefficient) for coefficient in coefficients]
    if coefficients and len(exponents) != len(coefficients):
        raise PolynomialConstructionError(
            "expected len(exponents) == len(coefficients); found %d != %d" % (
                len(exponents), len(coefficients)))

    if retain_coefficients is None:
        retain_coefficients = numpoly.get_options()["retain_coefficients"]
    if not retain_coefficients and coefficients:
        exponents, coefficients = remove_redundant_coefficients(exponents, coefficients)

    if isinstance(names, numpoly.ndpoly):
        names = names.names
    if isinstance(names, string_types):
        if exponents.shape[1] > 1:
            names = ["%s%d" % (names, idx)
                     for idx in range(exponents.shape[1])]
        else:
            names = [names]
    if names:
        if len(names) != exponents.shape[1]:
            raise PolynomialConstructionError(
                "Name length incompatible exponent length; "
                "len%s != %d" % (names, exponents.shape[1]))
        if sorted(set(names)) != sorted(names):
            raise PolynomialConstructionError(
                "Duplicate indeterminant names: %s" % names)

    if retain_names is None:
        retain_names = numpoly.get_options()["retain_names"]
    if not retain_names:
        exponents, names = remove_redundant_names(exponents, names)

    exponents_, count = numpy.unique(exponents, return_counts=True, axis=0)
    if numpy.any(count > 1):
        raise PolynomialConstructionError(
            "Duplicate exponent keys found: %s" % exponents_[count > 1][0])

    return exponents, coefficients, names


def remove_redundant_coefficients(exponents, coefficients):
    """
    Remove coefficients if they are redundant to the polynomial representation.

    Will always keep at least one coefficient.

    Args:
        exponents (numpy.ndarray):
            The exponents in an integer array with shape ``(N, D)``, where
            ``N`` is the number of terms in the polynomial sum and ``D`` is
            the number of dimensions.
        coefficients (Sequence[numpy.ndarray]):
            The polynomial coefficients. Must correspond to `exponents` by
            having the same length ``N``.

    Returns:
        Same as inputs, but with redundant exponent rows and coefficients
        removed. This corresponds to ``N`` being lowered.

    Examples:
        >>> exponents, coefficients = remove_redundant_coefficients(
        ...     [[0, 0], [0, 1]], [1, 2])
        >>> exponents
        array([[0, 0],
               [0, 1]])
        >>> coefficients
        [array(1), array(2)]
        >>> remove_redundant_coefficients(
        ...     [[0, 0], [0, 1]], [1, 0])
        (array([[0, 0]]), [array(1)])
        >>> remove_redundant_coefficients([[0]], [])
        (array([[0]]), [])

    """
    exponents = numpy.asarray(exponents)
    coefficients = [numpy.asarray(coefficient) for coefficient in coefficients]

    if not coefficients:
        assert exponents.shape == (1, 1)

    else:
        elements = list(zip(*[
            (exponent, coefficient)
            for exponent, coefficient in zip(exponents, coefficients)
            if numpy.any(coefficient) or not numpy.any(exponent)
        ]))
        exponents = numpy.asarray(elements[0], dtype=int)
        coefficients = list(elements[1])
    return exponents, coefficients


def remove_redundant_names(exponents, names):
    """
    Remove names if they are redundant to the polynomial representation.

    Will always keep at least one dimension.

    Args:
        exponents (numpy.ndarray):
            The exponents in an integer array with shape ``(N, D)``, where
            ``N`` is the number of terms in the polynomial sum and ``D`` is
            the number of dimensions.
        names (Sequence[str]):
            The indeterminant names, either as string names or as
            simple polynomials. Must correspond to the exponents by having
            length ``D``.

    Returns:
        Same as inputs, but with redundant exponent columns and names removed.
        This corresponds to ``D`` being lowered.

    Examples:
        >>> exponents, names = remove_redundant_names(
        ...     [[0, 0], [1, 1]], ["q0", "q1"])
        >>> exponents
        array([[0, 0],
               [1, 1]])
        >>> names
        ['q0', 'q1']
        >>> exponents, names = remove_redundant_names(
        ...     [[0, 0], [0, 1]], ["q0", "q1"])
        >>> exponents
        array([[0],
               [1]])
        >>> names
        ['q1']
        >>> exponents, names = remove_redundant_names(
        ...     [[0, 0]], ["q0", "q1"])
        >>> exponents
        array([[0]])
        >>> names
        ['q0']

    """
    exponents = numpy.asarray(exponents)

    indices = numpy.any(exponents != 0, 0)
    if not numpy.any(indices):
        indices[0] = True

    if names:
        names = list(names)
        names = numpy.array(names)[indices].tolist()
    exponents = exponents[:, indices]
    return exponents, names



"""Return (x1 == x2) element-wise."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.equal)
def equal(x1, x2, out=None, where=True, **kwargs):
    """
    Return (x1 == x2) element-wise.

    Args:
        x1, x2 (numpoly.ndpoly):
            Input arrays. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value.
            Note that if an uninitialized `out` array is created via the
            default ``out=None``, locations within it where the condition is
            False will remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (Union[numpy.ndarray, numpy.generic]):
            Output array, element-wise comparison of `x1` and `x2`. Typically
            of type bool, unless ``dtype=object`` is passed. This is a scalar
            if both `x1` and `x2` are scalars.

    Examples:
        >>> q0, q1, q2 = q0q1q2 = numpoly.variable(3)
        >>> numpoly.equal(q0q1q2, q0)
        array([ True, False, False])
        >>> numpoly.equal(q0q1q2, [q1, q1, q2])
        array([False,  True,  True])
        >>> numpoly.equal(q0, q1)
        False

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    if out is None:
        out = numpy.ones(x1.shape, dtype=bool)
    if not out.shape:
        return numpy.bool_(equal(x1.ravel(), x2.ravel(), out=out.ravel()).item())
    for coeff1, coeff2 in zip(x1.coefficients, x2.coefficients):
        out &= numpy.equal(coeff1, coeff2, where=where, **kwargs)
    return out



"""Return the truth value of (x1 > x2) element-wise."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.greater)
def greater(x1, x2, out=None, **kwargs):
    """
    Return the truth value of (x1 > x2) element-wise.

    Args:
        x1, x2 (numpoly.ndpoly):
            Input arrays. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Optional[numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        out (numpy.ndarray, scalar):
            Output array, element-wise comparison of `x1` and `x2`. Typically
            of type bool, unless ``dtype=object`` is passed. This is a scalar
            if both `x1` and `x2` are scalars.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.greater(3, 4)
        False
        >>> numpoly.greater(4*q0, 3*q0)
        True
        >>> numpoly.greater(q0, q1)
        False
        >>> numpoly.greater(q0**2, q0)
        True
        >>> numpoly.greater([1, q0, q0**2, q0**3], q1)
        array([False, False,  True,  True])
        >>> numpoly.greater(q0+1, q0-1)
        True
        >>> numpoly.greater(q0, q0)
        False

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    coefficients1 = x1.coefficients
    coefficients2 = x2.coefficients
    if out is None:
        out = numpy.greater(coefficients1[0], coefficients2[0], **kwargs)
    if not out.shape:
        return bool(greater(x1.ravel(), x2.ravel(), out=out.ravel()).item())

    options = numpoly.get_options()
    for idx in numpoly.glexsort(x1.exponents.T, graded=options["sort_graded"],
                                reverse=options["sort_reverse"]):

        indices = (coefficients1[idx] != 0) | (coefficients2[idx] != 0)
        indices &= coefficients1[idx] != coefficients2[idx]
        out[indices] = numpy.greater(
            coefficients1[idx], coefficients2[idx], **kwargs)[indices]
    return out



"""Return the indices of the elements that are non-zero."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.nonzero)
def nonzero(x, **kwargs):
    """
    Return the indices of the elements that are non-zero.

    Args:
        x (numpoly.ndpoly):
            Input array.

    Returns:
        indices (Union[bool, numpoly.ndarray]):
            Indices of elements that are non-zero.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([[3*q0, 0, 0],
        ...                            [0, 4*q1, 0],
        ...                            [5*q0+q1, 6*q0, 0]])
        >>> poly
        polynomial([[3*q0, 0, 0],
                    [0, 4*q1, 0],
                    [q1+5*q0, 6*q0, 0]])
        >>> numpoly.nonzero(poly)
        (array([0, 1, 2, 2]), array([0, 1, 0, 1]))
        >>> poly[numpoly.nonzero(poly)]
        polynomial([3*q0, 4*q1, q1+5*q0, 6*q0])

    """
    x = numpoly.aspolynomial(x)
    return numpy.nonzero(numpy.any(x.coefficients, axis=0))



"""Multi indices for monomial exponents."""
import numpy

from .glexindex import glexindex


def bindex(start, stop=None, dimensions=1, ordering="G", cross_truncation=1.):
    """
    Generate multi-indices for the monomial exponents.

    Args:
        start (Union[int, numpy.ndarray]):
            The lower order of the indices. If array of int, counts as lower
            bound for each axis.
        stop (Union[int, numpy.ndarray, None]):
            The maximum shape included. If omitted: stop <- start; start <- 0
            If int is provided, set as largest total order. If array of int,
            set as upper bound for each axis.
        dimensions (int):
            The number of dimensions in the expansion.
        ordering (str):
            Short hand for the criteria to sort the indices by.

            ``G``
                Graded sorting, meaning the indices are always sorted by the
                index sum. E.g. ``(2, 2, 2)`` has a sum of 6, and will
                therefore be consider larger than both ``(3, 1, 1)`` and
                ``(1, 1, 3)``.
            ``R``
                Lexicographical sorting, meaning that ``(1, 3)`` is
                considered bigger than ``(3, 1)``, instead of the opposite.
            ``I``
                The biggest values are in the front instead of the back.
        cross_truncation (float, Tuple[float, float]):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion. If two values are provided, first is low bound
            truncation, while the latter upper bound. If only one value, upper
            bound is assumed.

    Returns:
        list:
            Order list of indices.

    Examples:
        >>> numpoly.bindex(4).tolist()
        [[0], [1], [2], [3]]
        >>> numpoly.bindex(2, dimensions=2).tolist()
        [[0, 0], [0, 1], [1, 0]]
        >>> numpoly.bindex(start=2, stop=3, dimensions=2).tolist()
        [[0, 2], [1, 1], [2, 0]]
        >>> numpoly.bindex([1, 2, 3]).tolist()
        [[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 0, 2]]
        >>> numpoly.bindex([1, 2, 3], cross_truncation=numpy.inf).tolist()
        [[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 0, 2], [0, 1, 1], [0, 1, 2]]

    """
    ordering = ordering.upper()
    graded = "G" in ordering
    reverse = "R" not in ordering
    output = glexindex(
        start=start,
        stop=stop,
        dimensions=dimensions,
        cross_truncation=cross_truncation,
        graded=graded,
        reverse=reverse,
    )
    indices = slice(None, None, -1) if "I" in ordering else slice(None)
    return output[indices]



"""Evenly round to the given number of decimals."""
import numpy
import numpoly

from ..dispatch import implements, simple_dispatch


@implements(numpy.around, numpy.round)
def around(a, decimals=0, out=None):
    """
    Evenly round to the given number of decimals.

    Args:
        a (numpoly.ndpoly):
            Input data.
        decimals (Optional[int]):
            Number of decimal places to round to (default: 0). If decimals is
            negative, it specifies the number of positions to the left of the
            decimal point.
        out (Optional[numpy.ndarray]):
            Alternative output array in which to place the result. It must have
            the same shape as the expected output, but the type of the output
            values will be cast if necessary.

    Returns:
        (numpy.ndarray):
            An array of the same type as `a`, containing the rounded values.
            Unless `out` was specified, a new array is created.  A reference to
            the result is returned. The real and imaginary parts of complex
            numbers are rounded separately.  The result of rounding a float is
            a float.

    Examples:
        >>> q0 = numpoly.variable()
        >>> numpoly.around([0.37, 1.64*q0-2.45])
        polynomial([0.0, 2.0*q0-2.0])
        >>> numpoly.around([0.37, 1.64*q0-23.45], decimals=1)
        polynomial([0.4, 1.6*q0-23.4])
        >>> numpoly.around([0.37, 1.64*q0-23.45], decimals=-1)
        polynomial([0.0, -20.0])

    """
    return simple_dispatch(
        numpy_func=numpy.around,
        inputs=(a,),
        out=out,
        decimals=decimals,
    )



"""Return an array of zeros with the same shape and type as a given array."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.zeros_like)
def zeros_like(a, dtype=None, order='K', subok=True, shape=None):
    """
    Return an array of zeros with the same shape and type as a given array.

    Args:
        a (numpoly.ndpoly):
            The shape and data-type of `a` define these same attributes of
            the returned array.
        dtype (Optional[numpy.dtype]):
            Overrides the data type of the result.
        order (str):
            Overrides the memory layout of the result. 'C' means C-order,
            'F' means F-order, 'A' means 'F' if `a` is Fortran contiguous,
            'C' otherwise. 'K' means match the layout of `a` as closely
            as possible.
        subok (bool):
            If True, then the newly created array will use the sub-class
            type of 'a', otherwise it will be a base-class array. Defaults
            to True.
        shape (Optional[Sequence[int]]):
            Overrides the shape of the result. If order='K' and the number of
            dimensions is unchanged, will try to keep order, otherwise,
            order='C' is implied.

    Returns:
        (numpoly.ndpoly):
            Array of zeros with the same shape and type as `a`.

    Examples:
        >>> poly = numpoly.monomial(3)
        >>> poly
        polynomial([1, q0, q0**2])
        >>> numpoly.zeros_like(poly)
        polynomial([0, 0, 0])

    """
    del subok
    if not isinstance(a, numpy.ndarray):
        a = numpoly.polynomial(a)
    if shape is None:
        shape = a.shape
    if dtype is None:
        dtype = a.dtype
    if order in ("A", "K"):
        order = "F" if a.flags["F_CONTIGUOUS"] else "C"
    return numpoly.polynomial(numpy.zeros(shape, dtype=dtype, order=order))



"""Sort keys lexicographically."""
import numpy


def glexsort(keys, graded=False, reverse=False):
    """
    Sort keys graded lexicographically.

    Same as ``numpy.lexsort``, but also support graded reverse
    lexicographical ordering.

    Args:
        keys (Sequence(Sequence[int])):
            Values to sort.
        graded (bool):
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``(2, 2, 2)`` has a sum of 6, and will therefore be
            consider larger than both ``(3, 1, 1)`` and ``(1, 1, 3)``.
        reverse (bool):
            Reverse lexicographical sorting meaning that ``(1, 3)`` is
            considered smaller than ``(3, 1)``, instead of the opposite.

    Returns:
        (numpy.ndarray):
            Array of indices that sort the keys along the specified axis.

    Examples:
        >>> indices = numpy.array([[0, 0, 0, 1, 2, 1],
        ...                        [1, 2, 0, 0, 0, 1]])
        >>> indices[:, numpy.lexsort(indices)]
        array([[0, 1, 2, 0, 1, 0],
               [0, 0, 0, 1, 1, 2]])
        >>> indices[:, numpoly.glexsort(indices)]
        array([[0, 1, 2, 0, 1, 0],
               [0, 0, 0, 1, 1, 2]])
        >>> indices[:, numpoly.glexsort(indices, reverse=True)]
        array([[0, 0, 0, 1, 1, 2],
               [0, 1, 2, 0, 1, 0]])
        >>> indices[:, numpoly.glexsort(indices, graded=True)]
        array([[0, 1, 0, 2, 1, 0],
               [0, 0, 1, 0, 1, 2]])
        >>> indices[:, numpoly.glexsort(indices, graded=True, reverse=True)]
        array([[0, 0, 1, 0, 1, 2],
               [0, 1, 0, 2, 1, 0]])
        >>> indices = numpy.array([4, 5, 6, 3, 2, 1])
        >>> indices[numpoly.glexsort(indices)]
        array([1, 2, 3, 4, 5, 6])

    """
    keys = numpy.atleast_2d(keys)

    if reverse:
        keys = keys[::-1]

    indices = numpy.array(numpy.lexsort(keys))
    if graded:
        indices = indices[numpy.argsort(
            numpy.sum(keys[:, indices], axis=0))].T
    return indices



"""Return a string representation of the data in an array."""
import numpy
import numpoly

from ..dispatch import implements
from .array_repr import to_string


@implements(numpy.array_str)
def array_str(a, max_line_width=None, precision=None, suppress_small=None):
    """
    Return a string representation of the data in an array.

    The data in the array is returned as a single string.  This function is
    similar to `array_repr`, the difference being that `array_repr` also
    returns information on the kind of array and its data type.

    Args:
        a (numpoly.ndpoly):
            Input array.
        max_line_width (Optional[int]):
            Inserts newlines if text is longer than `max_line_width`. Defaults
            to ``numpy.get_printoptions()['linewidth']``.
        precision (Optional[int]):
            Floating point precision. Defaults to
            ``numpy.get_printoptions()['precision']``.
        suppress_small (Optional[bool]):
            Represent numbers "very close" to zero as zero; default is False.
            Very close is defined by precision: if the precision is 8, e.g.,
            numbers smaller (in absolute value) than 5e-9 are represented as
            zero. Defaults to ``numpy.get_printoptions()['suppress']``.

    Returns:
        (str):
            The string representation of an array.

    Examples:
        >>> q0 = numpoly.variable()
        >>> numpoly.array_str(numpoly.polynomial([1, q0]))
        '[1 q0]'
        >>> numpoly.array_str(numpoly.polynomial([]))
        '[]'
        >>> numpoly.array_str(
        ...     numpoly.polynomial([1e-6, 4e-7*q0, 2*q0, 3]),
        ...     precision=4,
        ...     suppress_small=True,
        ... )
        '[0.0 0.0 2.0*q0 3.0]'

    """
    a = numpoly.aspolynomial(a)
    a = to_string(a, precision=precision, suppress_small=suppress_small)
    return numpy.array2string(
        numpy.array(a),
        max_line_width=max_line_width,
        separator=" ",
        formatter={"all": str},
        prefix="",
        suffix="",
    )



"""Return a new array of given shape and type, filled with `fill_value`."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.full)
def full(shape, fill_value, dtype=None, order="C"):
    """
    Return a new array of given shape and type, filled with `fill_value`.

    Args:
        shape (int, Sequence[int]):
            Shape of the new array, e.g., ``(2, 3)`` or ``2``.
        fill_value (numpoly.ndpoly):
            Fill value. Must be broadcast compatible with `shape`.
        dtype : data-type, optional
            The desired data-type for the array  The default, None, means
            inherit from `fill_value`.
        order : {'C', 'F'}, optional
            Whether to store multidimensional data in C- or Fortran-contiguous
            (row- or column-wise) order in memory.

    Returns:
        out : ndarray
            Array of `fill_value` with the given shape, dtype, and order.

    Examples:
        >>> numpoly.full((2, 4), 4)
        polynomial([[4, 4, 4, 4],
                    [4, 4, 4, 4]])
        >>> q0 = numpoly.variable()
        >>> numpoly.full(3, q0**2-1)
        polynomial([q0**2-1, q0**2-1, q0**2-1])

    """
    fill_value = numpoly.aspolynomial(fill_value)
    if dtype is None:
        dtype = fill_value.dtype
    out = numpoly.ndpoly(
        exponents=fill_value.exponents,
        shape=shape,
        names=fill_value.indeterminants,
        dtype=dtype,
        order=order,
    )
    for key in fill_value.keys:
        out[key] = fill_value[key]
    return out



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['numpoly',
 'numpoly.array_function',
 'numpoly.construct',
 'numpoly.poly_function',
 'numpoly.poly_function.divide',
 'numpoly.utils']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.16,<2.0', 'six']

setup_kwargs = {
    'name': 'numpoly',
    'version': '1.1.2',
    'description': 'Polynomials as a numpy datatype',
    'long_description': ".. image:: https://github.com/jonathf/numpoly/raw/master/docs/.static/numpoly_logo.svg\n   :height: 200 px\n   :width: 200 px\n   :align: center\n\n|circleci| |codecov| |pypi| |readthedocs|\n\n.. |circleci| image:: https://circleci.com/gh/jonathf/numpoly/tree/master.svg?style=shield\n    :target: https://circleci.com/gh/jonathf/numpoly/tree/master\n.. |codecov| image:: https://codecov.io/gh/jonathf/numpoly/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/jonathf/numpoly\n.. |pypi| image:: https://badge.fury.io/py/numpoly.svg\n    :target: https://badge.fury.io/py/numpoly\n.. |readthedocs| image:: https://readthedocs.org/projects/numpoly/badge/?version=master\n    :target: http://numpoly.readthedocs.io/en/master/?badge=master\n\nNumpoly is a generic library for creating, manipulating and evaluating\narrays of polynomials based on ``numpy.ndarray`` objects.\n\n.. contents:: Table of Contents:\n\nFeature Overview\n----------------\n\n* Intuitive interface for users experienced with ``numpy``, as the library\n  provides a high level of compatibility with the ``numpy.ndarray``, including\n  fancy indexing, broadcasting, ``numpy.dtype``, vectorized operations to name\n  a few.\n* Computationally fast evaluations of lots of functionality inherent from\n  ``numpy``.\n* Vectorized polynomial evaluation.\n* Support for arbitrary number of dimensions.\n* Native support for lots of ``numpy.<name>`` functions using ``numpy``'s\n  compatibility layer (which also exists as ``numpoly.<name>``\n  equivalents).\n* Support for polynomial division through the operators ``/``, ``%`` and\n  ``divmod``.\n* Extra polynomial specific attributes exposed on the polynomial objects like\n  ``poly.exponents``, ``poly.coefficients``, ``poly.indeterminants`` etc.\n* Polynomial derivation through functions like ``numpoly.derivative``,\n  ``numpoly.gradient``, ``numpoly.hessian`` etc.\n* Decompose polynomial sums into vector of addends using ``numpoly.decompose``.\n* Variable substitution through ``numpoly.call``.\n\nInstallation\n------------\n\nInstallation should be straight forward:\n\n.. code-block:: bash\n\n    pip install numpoly\n\nExample Usage\n-------------\n\nConstructing polynomial is typically done using one of the available\nconstructors:\n\n.. code-block:: python\n\n    >>> import numpoly\n    >>> numpoly.monomial(start=0, stop=3, dimensions=2)\n    polynomial([1, q0, q0**2, q1, q0*q1, q1**2])\n\nIt is also possible to construct your own from symbols together with\n`numpy <https://python.org>`_:\n\n.. code-block:: python\n\n    >>> import numpy\n    >>> q0, q1 = numpoly.variable(2)\n    >>> numpoly.polynomial([1, q0**2-1, q0*q1, q1**2-1])\n    polynomial([1, q0**2-1, q0*q1, q1**2-1])\n\nOr in combination with numpy objects using various arithmetics:\n\n.. code-block:: python\n\n    >>> q0**numpy.arange(4)-q1**numpy.arange(3, -1, -1)\n    polynomial([-q1**3+1, -q1**2+q0, q0**2-q1, q0**3-1])\n\nThe constructed polynomials can be evaluated as needed:\n\n.. code-block:: python\n\n    >>> poly = 3*q0+2*q1+1\n    >>> poly(q0=q1, q1=[1, 2, 3])\n    polynomial([3*q1+3, 3*q1+5, 3*q1+7])\n\nOr manipulated using various numpy functions:\n\n.. code-block:: python\n\n    >>> numpy.reshape(q0**numpy.arange(4), (2, 2))\n    polynomial([[1, q0],\n                [q0**2, q0**3]])\n    >>> numpy.sum(numpoly.monomial(13)[::3])\n    polynomial(q0**12+q0**9+q0**6+q0**3+1)\n",
    'author': 'Jonathan Feinberg',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jonathf/numpoly',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
}


setup(**setup_kwargs)



"""Test whether all array elements along a given axis evaluate to True."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.all)
def all(a, axis=None, out=None, keepdims=False, **kwargs):
    """
    Test whether all array elements along a given axis evaluate to True.

    Args:
        a (numpoly.ndpoly):
            Input array or object that can be converted to an array.
        axis (Union[int, Tuple[int], None]):
            Axis or axes along which a logical AND reduction is performed. The
            default (`axis` = `None`) is to perform a logical AND over all the
            dimensions of the input array. `axis` may be negative, in which
            case it counts from the last to the first axis. If this is a tuple
            of ints, a reduction is performed on multiple axes, instead of
            a single axis or all the axes as before.
        out (Optional[numpy.ndarray]):
            Alternate output array in which to place the result. It must have
            the same shape as the expected output and its type is preserved
            (e.g., if ``dtype(out)`` is float, the result will consist of 0.0's
            and 1.0's).
        keepdims (Optional[bool]):
            If this is set to True, the axes which are reduced are left
            in the result as dimensions with size one. With this option,
            the result will broadcast correctly against the input array.

    Returns:
        any (Union[bool, numpy.ndarray]):
            A new boolean or array is returned unless `out` is specified, in
            which case a reference to `out` is returned.

    Examples:
        >>> q0 = numpoly.variable()
        >>> numpoly.all(q0)
        True
        >>> numpoly.all(0*q0)
        False
        >>> numpoly.all([1, q0, 0])
        False
        >>> numpoly.all([[True*q0, False], [True, True]], axis=0)
        array([ True, False])

    """
    a = numpoly.aspolynomial(a)
    coefficients = numpy.any(a.coefficients, 0).astype(bool)
    out = numpy.all(
        coefficients, axis=axis, out=out, keepdims=keepdims)
    return out



"""Stack arrays in sequence vertically (row wise)."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.vstack)
def vstack(tup):
    """
    Stack arrays in sequence vertically (row wise).

    This is equivalent to concatenation along the first axis after 1-D arrays
    of shape `(N,)` have been reshaped to `(1,N)`. Rebuilds arrays divided by
    `vsplit`.

    This function makes most sense for arrays with up to 3 dimensions. For
    instance, for pixel-data with a height (first axis), width (second axis),
    and r/g/b channels (third axis). The functions `concatenate`, `stack` and
    `block` provide more general stacking and concatenation operations.

    Args:
        tup (Sequence[numpoly.ndpoly]):
            The arrays must have the same shape along all but the first axis.
            1-D arrays must have the same length.

    Returns:
        (numpoly.ndpoly):
            The array formed by stacking the given arrays, will be at least
            2-D.

    Examples:
        >>> poly1 = numpoly.variable(3)
        >>> const1 = numpoly.polynomial([1, 2, 3])
        >>> numpoly.vstack([poly1, const1])
        polynomial([[q0, q1, q2],
                    [1, 2, 3]])
        >>> const2 = numpoly.polynomial([[1], [2], [3]])
        >>> poly2 = poly1.reshape(3, 1)
        >>> numpoly.vstack([const2, poly2])
        polynomial([[1],
                    [2],
                    [3],
                    [q0],
                    [q1],
                    [q2]])

    """
    arrays = numpoly.align_exponents(*tup)
    arrays = numpoly.align_dtype(*arrays)
    result = numpy.vstack([array.values for array in arrays])
    return numpoly.aspolynomial(result, names=arrays[0].indeterminants)



"""Return a new array of given shape and type, filled with zeros."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.zeros)
def zeros(shape, dtype=float, order="C"):
    """
    Return a new array of given shape and type, filled with zeros.

    Args:
        shape : int or tuple of ints
            Shape of the new array, e.g., ``(2, 3)`` or ``2``.
        dtype : data-type, optional
            The desired data-type for the array, e.g., `numpy.int8`.  Default is
            `numpy.float64`.
        order : {'C', 'F'}, optional, default: 'C'
            Whether to store multi-dimensional data in row-major
            (C-style) or column-major (Fortran-style) order in
            memory.

    Returns:
        (numpoly.ndpoly):
            Array of zeros with the given shape, dtype, and order.

    Examples:
        >>> numpoly.zeros(5)
        polynomial([0.0, 0.0, 0.0, 0.0, 0.0])

    """
    return numpoly.polynomial(numpy.zeros(shape, dtype=dtype, order=order))
